import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.Reducer;

/**
 * Reduce class
 * @author SergeiL
 */
public class ClaraGroupReducer extends Reducer<LongWritable, Text, LongWritable, Text>
{
	@Override
	protected void reduce(LongWritable key, Iterable<Text> values, Context context) throws IOException, InterruptedException
	{
		List<Point> medoids = handleData(parseData(values), context);
		String result = ":";
		
		for (Point p : medoids)
		{
			result += p.getX() + ";" + p.getY() + ",";
		}
		
		result = result.substring(0, result.length() - 1);
		context.write(key, new Text(result));
	}
	
	public List<Point> parseData(Iterable<Text> values) {
		List<Point> result = new ArrayList<Point>();
		String[] valueArr;
		
		for (Text value : values)
		{
			valueArr = value.toString().split(";");
			result.add(new Point(Integer.parseInt(valueArr[0]), Integer.parseInt(valueArr[1])));
		}
		
		return result;
	}
	
	public List<Point> handleData(List<Point> value, Context context) {
		int sampleSize = context.getConfiguration().getInt("DATA_SAMPLE_SIZE", 0);
		int medoidsNum = context.getConfiguration().getInt("NUM_OF_MEDOIDS", 0);
		
		ArrayList<Integer> selectedIdxs = getRandomNumbers(sampleSize, value.size() - 1);
		List<Point> dataChunk = new ArrayList<Point>();	
		ArrayList<Integer> medoidIds;
		List<Point> medoids = new ArrayList<Point>();		
		int i;
		int j;
		int iterNum;
		List<Point> prevResult = null;
		List<Point> result = null;
		
		for (i = 0; i < selectedIdxs.size(); i++) dataChunk.add(value.get(selectedIdxs.get(i)));
		medoidIds = getRandomNumbers(medoidsNum, dataChunk.size() - 1);
		for (i = 0; i < medoidIds.size(); i++) medoids.add(dataChunk.get(medoidIds.get(i)));
		iterNum = dataChunk.size() * 3 / 4;
		
		for (i = 0; i < iterNum; i++)
		{
			result = initPointMedoids(medoids, dataChunk/*, false*/);
			if(prevResult != null && checkMedoids(prevResult, result)) 
			{
				for (j = 0; j < result.size(); j++) result.get(j).clearAssocPoints();
				return result;
			}
			prevResult = result;
		}
		
		for (i = 0; i < result.size(); i++) result.get(i).clearAssocPoints();
		return result;
	}
	
	private boolean checkMedoids(List<Point> prev, List<Point> current)
	{
		double prevSum = 0;
		double currentSum = 0;
		
		for (int i = 0; i < prev.size(); i++) 
		{
			prevSum += prev.get(i).getClusterCost();
			currentSum += current.get(i).getClusterCost();
		}
			
		return (prevSum == currentSum);
	}
	
	private List<Point> initPointMedoids(List<Point> medoids, List<Point> points/*, boolean whole*/)
	{
		double minDist;
		double distance;
		Point medoid;
		Point selectedMedoid = null;
		Point p;
		List<Point> result = new ArrayList<Point>();
		int i;
		int j;
		for (i = 0; i < medoids.size(); i++) medoids.get(i).clearAssocPoints();
		
		for (i = 0; i < points.size(); i++)
		{
			minDist = Double.MAX_VALUE;
			p = points.get(i);			
			
			for (j = 0; j < medoids.size(); j++)
			{
				medoid = medoids.get(j);
				distance = calcDistance(p.getX(), medoid.getX(), p.getY(), medoid.getY());
				
				if(distance < minDist) 
				{
					minDist = distance;
					selectedMedoid = medoid;
				}
			}
			
			selectedMedoid.addAssocPoint(p);
			selectedMedoid.incrClusterCost(minDist);
		}
		
		for (i = 0; i < medoids.size(); i++) result.add(recalcMedoidPos(medoids.get(i)));		
		return result;
	}
	
	private Point recalcMedoidPos(Point medoid)
	{
		Point result = medoid;
		List<Point> clusterPoints = medoid.getAssocPoints();
		double minCost = medoid.getClusterCost();
		double cost;
		Point p;
		int i;
		int j;
		
		for (i = 0; i < clusterPoints.size(); i++) 
		{
			medoid = clusterPoints.get(i);	
			cost = 0;
			
			for (j = 0; j < clusterPoints.size(); j++) 
			{
				p = clusterPoints.get(j);	
				cost += calcDistance(p.getX(), medoid.getX(), p.getY(), medoid.getY());
			}
			
			if(cost < minCost)
			{
				minCost = cost;
				result = medoid;
				result.setClusterCost(minCost);
			}
		}
		
		return result;
	}
	
	/**
	 * Calculate distance
	 * @param x1 - first x coordinate
	 * @param x2 - second x coordinate
	 * @param y1 - first y coordinate
	 * @param y2 - second y coordinate
	 * @return double
	 */
	private double calcDistance(long x1, long x2, long y1, long y2) {
		// Double.parseDouble(new
		// DecimalFormat("#.##").format(unformattedNumber));
		return Math.sqrt(Math.pow((x2 - x1), 2) + Math.pow((y2 - y1), 2));
	}
	
	/**
	 * Get random numbers
	 * @param size - list size
	 * @param max - max number
	 * @return LinkedList<Integer>
	 */
	private static ArrayList<Integer> getRandomNumbers(int size, int max) {
		Random rng = new Random(); // Ideally just create one instance globally
		// Note: use LinkedHashSet to maintain insertion order
		ArrayList<Integer> result = new ArrayList<Integer>();
		if(size > max) throw new Error("Size must be lesser than max: " + size + " < " + max);

		while (result.size() < size) {
			// long next = (long) (rng.nextDouble() * max);
			int next = rng.nextInt(max) + 1;
			// As we're adding to a set, this will automatically do a
			// containment check
			if(!result.contains(next)) result.add(next);
		}

		return result;
	}
}