import java.io.IOException;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;

import org.apache.spark.api.java.*;
import org.apache.spark.api.java.function.*;

import scala.Tuple2;

/**
 * Spark parallel naive bayesian class
 * @author SergeiL
 */
public class Main {
	// Broadcasted train data
	private static JavaRDD<DataField> mTrainData;

	/**
	 * Main function
	 * @param args
	 * args[0] - train data path 
	 * args[1] - test data path
	 * @throws IOException 
	 */
	public static void main(String[] args) throws IOException {
		double start = System.currentTimeMillis();
		String testDataPath = "testData.csv";
		String trainDataPath = "trainData.csv";
		int procNum = 1;
		
		if(args != null && args.length > 0 && args[0] != null) testDataPath = args[0];
		if(args != null && args.length > 1 && args[1] != null) procNum = Integer.parseInt(args[1]);
		if(args != null && args.length > 2 && args[2] != null) trainDataPath = args[2];
		
		DateFormat dateFormat = new SimpleDateFormat("yyyy_MM_dd_HH_mm_ss");
		Date date = new Date();
		JavaSparkContext sc = new JavaSparkContext("spark://laada-spark-master:7077", "Spark PnBm App", "/root/spark-0.8.0-incubating", "SparkPnBm.jar");
		mTrainData = sc.textFile("hdfs://laada-spark-master:54310/user/hduser/SparkPnBm/" + trainDataPath).map(new ParseData());
		JavaRDD<DataField> testData = sc.textFile("hdfs://laada-spark-master:54310/user/hduser/SparkPnBm/" + testDataPath, procNum).map(new ParseData());
		JavaPairRDD<String, String> mapData = testData.map(new MapElement(mTrainData.collect()));		
		mapData.saveAsTextFile("hdfs://laada-spark-master:54310/user/hduser/SparkPnBm/result/" + dateFormat.format(date));
		System.out.println("-----RESULT TIME: " + ((System.currentTimeMillis() - start) / 1000));
	}
}

/**
 * Map class to process data
 * @author SergeiL
 */
@SuppressWarnings("serial")
class MapElement extends PairFunction<DataField, String, String> {
	// Train data
	private final List<DataField> mTrainData;

	/**
	 * Constructor
	 * @param values - train data values
	 */
	MapElement(List<DataField> values) {
		mTrainData = values;
	}

	/**
	 * Process data function
	 */
	public Tuple2<String, String> call(DataField value) {
		double pHeight;
		double pWeight;
		double pFootSize;
		double pMale = 0;
		double pFemale = 0;

		for (DataField element : mTrainData) {
			pHeight = (1 / Math.sqrt(2 * Math.PI
					* element.getVarianceHeight()))
					* Math.exp(-Math.pow(
							value.getHeight() - element.getMeanHeight(),
							2) / (2 * element.getVarianceHeight()));
			pWeight = (1 / Math.sqrt(2 * Math.PI
					* element.getVarianceWeight()))
					* Math.exp(-Math.pow(
							value.getWeight() - element.getMeanWeight(),
							2) / (2 * element.getVarianceWeight()));
			pFootSize = (1 / Math.sqrt(2 * Math.PI
					* element.getVarianceFootSize()))
					* Math.exp(-Math.pow(
							value.getFootSize()
									- element.getMeanFootSize(), 2)
							/ (2 * element.getVarianceFootSize()));
			
			if (pFemale == 0) pFemale = 0.5 * pHeight * pWeight * pFootSize;
			else pMale = 0.5 * pHeight * pWeight * pFootSize;
		}

		// Decide male or female
		if (pMale > pFemale) return new Tuple2<String, String>(value.getSex(), "Male");
		else return new Tuple2<String, String>(value.getSex(), "Female");
	}
}

/**
 * Parse data class
 * @author SergeiL
 */
@SuppressWarnings("serial")
class ParseData extends Function<String, DataField> {
	
	/**
	 * Process data
	 */
	public DataField call(String value) {
		String[] valueArr = value.split(";");
		if(valueArr.length == 4) return new DataField(valueArr[0], Double.parseDouble(valueArr[1]), Double.parseDouble(valueArr[2]), Double.parseDouble(valueArr[3]));
		return new DataField(valueArr[0], Double.parseDouble(valueArr[1]), Double.parseDouble(valueArr[2]), Double.parseDouble(valueArr[3]), Double.parseDouble(valueArr[4]), Double.parseDouble(valueArr[5]), Double.parseDouble(valueArr[6]));
	}
}